// for in-program strings
#include "Flash.h"

#include "Fishino.h"

#include "Arduino.h"
#include "SPI.h"

enum Command
{
	CMD_NOP,			// no-op command								0
	CMD_RESET,			// chip reset									1
	CMD_TEST,			// test - returns given strings parameters		2
	CMD_SETUART,		// setup spi interface parameters				3
	CMD_GETFWVERSION,	// get firmware version							4


	CMD_QUERYMODE,		// query chip mode (AP, STA or AP+STA)			5
	CMD_SETMODE,		// set chip mode								6
	CMD_LISTAP,			// get list of available AP						7
	CMD_JOINAP,			// join an access point							8
	CMD_QUITAP,			// quit access point							9
	CMD_QUERYAP,		// query info of joined AP						10

	CMD_SETSTAIP,		// set station IP								11
	CMD_GETSTAIP,		// get station IP								12
	CMD_SETSTANM,		// set station Netmask							13
	CMD_GETSTANM,		// get station Netmask							14
	CMD_SETSTAGW,		// set station gateway							15
	CMD_GETSTAGW,		// get station gatway							16
	CMD_SETSTADNS,		// set station DNS								17
	CMD_GETSTADNS,		// get station DNS								18
	CMD_SETSTAMAC,		// set station MAC								19
	CMD_GETSTAMAC,		// get station MAC								20
	CMD_ENABLESTADHCP,	// enable station DHCP							21
	CMD_DISABLESTADHCP,	// disable station DHCP							22

	CMD_SETAPIP,		// set AP IP									23
	CMD_GETAPIP,		// get AP IP									24
	CMD_SETAPNM,		// set AP Netmask								25
	CMD_GETAPNM,		// get AP Netmask								26
	CMD_SETAPGW,		// set AP gateway								27
	CMD_GETAPGW,		// get AP gatway								28
	CMD_SETAPDNS,		// set AP DNS									29
	CMD_GETAPDNS,		// get AP DNS									30
	CMD_SETAPMAC,		// set AP MAC									31
	CMD_GETAPMAC,		// get AP MAC									32
	CMD_ENABLEAPDHCP,	// enable AP DHCP								33
	CMD_DISABLEAPDHCP,	// disable AP DHCP								34

	CMD_CONNSTATUS,		// get tcp connection status					35
	CMD_CONNECT,		// starts a TCP or UDP connection				36
	CMD_CLOSE,			// close connection								37
	CMD_SEND,			// send data to socket							38
	CMD_AVAIL,			// get number of available data on a socket		39
	CMD_READ,			// read data from socket						40
	CMD_PEEK,			// peeks a byte of data from socket, if any		41
	CMD_FLUSH,			// flushes socket data							42
	
	CMD_STARTSERVER,	// start server									43
	CMD_STOPSERVER,		// stop server									44
	CMD_POLLSERVER,		// poll server for sockets with data			45

	CMD_APSETCONFIG,	// sets ap config (SSID, PASS, CHANNEL, HIDDEN)	46
	CMD_APGETCONFIG,	// gets ap config								47
	
	CMD_SETAPIPINFO,	// sets full AP IP info (IP, GW and NM)			48

	CMD_QUERYSTADHCP,	// queries status of station DHCP client		49
	CMD_QUERYAPDHCPS,	// queries statis of AP DHCP server				50
	CMD_SETAPDHCPRANGE,	// sets AP DHCP range							51

	CMD_SETSTACONFIG,	// sets station configuration data				52
	CMD_GETSTACONFIG,	// get station configuration data				53
	CMD_STACONNECT,		// connect station using stored parameters		54
	CMD_STADISCONNECT,	// disconnect station (duplicate of CMD_QUITAP)	55

	CMD_PINMODE,		// pinMode for extra I/O on ESP module			56
	CMD_DIGITALREAD,	// digital read for extra I/O on ESP module		57
	CMD_DIGITALWRITE,	// digital write for extra I/O on ESP module	58
	CMD_ANALOGREAD,		// analog read									59

	CMD_GETPHYMODE,		// get WiFi physical connection mode			60
	CMD_SETPHYMODE,		// set WiFi physical connection mode			61

};

typedef enum
{
	UART_OK,
	UART_BAD_PARAM,
	UART_ERROR
	
} UART_STATUS;


typedef enum
{
	AP_SCAN_STARTED,
	AP_SCAN_DONE,
	AP_SCAN_ERROR,
	AP_BAD_MODE,
	AP_REQUEST_TIMEOUT,
	AP_BAD_DATA,

} AP_SCAN_STATUS; 

typedef enum
{
	AP_CONFIG_OK,
	AP_CONFIG_FAILED,
	AP_CONFIG_BAD_MODE,
	AP_CONFIG_BAD_DATA

} AP_CONFIG_STATUS;

typedef enum
{
	DHCP_ERROR,
	DHCP_OK
	
} DHCP_STATUS;

typedef enum
{
	MAC_ERROR,
	MAC_OK

} MAC_STATUS;

typedef enum
{
	IP_ERROR,
	IP_OK
	
} IP_STATUS;

typedef enum
{
	COMMAND_OK,
	COMMAND_ERROR,
	COMMAND_BAD_DATA,
	COMMAND_UART_TIMEOUT,
	COMMAND_TIMEOUT,
	
} COMMAND_STATUS;

typedef enum
{
	CONNECTION_BAD_DATA = 256,
	CONNECTION_DNSFAIL,
	CONNECTION_TIMEOUT,
	CONNECTION_SOCKET_ERROR,
	CONNECTION_ERROR,
	CONNECTION_UART_TIMEOUT,

} CONNECTION_STATUS;

typedef enum
{
	DISCONNECTION_BAD_DATA,
	DISCONNECTION_TIMEOUT,
	DISCONNECTION_SOCKET_ERROR,
	DISCONNECTION_ERROR,
	DISCONNECTION_UART_TIMEOUT,
	DISCONNECTION_DISCONNECT_ERROR,
	DISCONNECTION_DISCONNECTED

} DISCONNECTION_STATUS;

typedef enum
{
	SEND_OK,
	SEND_ERROR,
	SEND_TIMEOUT,
	SEND_BAD_SOCKET,
	SEND_BAD_DATA,
	SEND_UART_TIMEOUT,
	SEND_TOO_BIG,
	
} SEND_STATUS;

typedef enum
{
	RECEIVE_OK,
	RECEIVE_ERROR,
	RECEIVE_TIMEOUT,
	RECEIVE_BAD_SOCKET,
	RECEIVE_BAD_DATA,
	RECEIVE_UART_TIMEOUT,
	RECEIVE_OUT_OF_MEMORY,
	
} RECEIVE_STATUS;

typedef enum
{
	TCPSERVER_LISTENING,
	TCPSERVER_OFF,
	TCPSERVER_ERROR,
	TCPSERVER_ALREADY,
	TCPSERVER_BAD_DATA,
	TCPSERVER_UART_TIMEOUT,
	
} TCPSERVER_STATUS;

// frees APINFO array
void FishinoClass::freeApInfo(void)
{
	if(_apInfo)
	{
		for(int i = 0; i < _numAp; i++)
			if(_apInfo[i].ssid)
				free(_apInfo[i].ssid);
		free(_apInfo);
		_apInfo = NULL;
	}
	_numAp = 0;
}

// find an APINFO given SSID
APINFO const *FishinoClass::findApInfo(const char *ssid)
{
	if(!ssid)
		return NULL;
	for(int i = 0; i < _numAp; i++)
		if(!strcmp(ssid, _apInfo[i].ssid))
			return &_apInfo[i];
	return NULL;
}

// low-level i/o routines
void FishinoClass::spiSelect(bool s)
{
	if(s)
	{
		::digitalWrite(ESP8266_CS, LOW);
		delayMicroseconds(20);
	}
	else
	{
		delayMicroseconds(20);
		::digitalWrite(ESP8266_CS, HIGH);
	}
}

// wait for busy handshake, or 100 useconds
// which comes first
void FishinoClass::spiWaitBusy(void)
{
	uint32_t t = micros() + 100;
	
	while(!::digitalRead(ESP8266_HANDSHAKE))
		if(micros() >= t)
			break;
}


// check if data is available from spi
bool FishinoClass::spiDataAvailable(void)
{
	return !::digitalRead(ESP8266_HANDSHAKE);
}

// wait for spi data available, timeout in milliseconds
bool FishinoClass::spiWaitDataAvailable(uint16_t tOut)
{
	uint32_t t = millis() + tOut;
	while(::digitalRead(ESP8266_HANDSHAKE))
		if(millis() >= t)
			break;
	return !::digitalRead(ESP8266_HANDSHAKE);
}


// send/receive a chunk of 1..64 bytes
// beware, ESP MUST know data size in advance
bool FishinoClass::spiSend(const uint8_t *b, uint8_t size)
{
	// wait for ESP to be ready, or timeout (2 sec should be enough)
	uint32_t tout = millis() + 2000;
	while(::digitalRead(ESP8266_HANDSHAKE))
		if(millis() > tout)
		{
			Serial << F("Fishino busy sending byte\n");
			return false;
		}
	
	// send byte through spi
	SPI.transfer(0x02);		// write command
	SPI.transfer(0xff);		// dummy address (sigh)
	for(uint8_t i = 0; i < size; i++)
		SPI.transfer(*b++);

	// wait for ESP to be busy processing byte
	// so we give time for ESP interupt handler to start
	spiWaitBusy();
	return true;
}

bool FishinoClass::spiReceive(uint8_t *b, uint8_t size)
{
	// read byte, timeout of 200 mSec
	unsigned long t = millis() + 200;
	while(::digitalRead(ESP8266_HANDSHAKE))
		if(millis() > t)
		{
			Serial << F("OOPS... ESP still busy\n");
			return false;
		}

	// send byte through spi
	SPI.transfer(0x03);			// read command
	SPI.transfer(0x00);			// dummy address (sigh)

	for(uint8_t i = 0; i < size; i++)
		*b++ = SPI.transfer(0x00);		// receive byte

	// wait for ESP to be busy processing byte
	// so we give time for ESP interupt handler to start
	spiWaitBusy();
	return true;
}

// drops data from server (like spiReceive() but don't save it
bool FishinoClass::spiDrop(uint8_t size)
{
	// read byte, timeout of 200 mSec
	unsigned long t = millis() + 200;
	while(::digitalRead(ESP8266_HANDSHAKE))
		if(millis() > t)
		{
			Serial << F("OOPS... ESP still busy\n");
			return false;
		}

	// send byte through spi
	SPI.transfer(0x03);			// read command
	SPI.transfer(0x00);			// dummy address (sigh)
	for(int i = 0; i < size; i++)
		SPI.transfer(0x00);		// receive byte and drop it

	// wait for ESP to be busy processing byte
	// so we give time for ESP interupt handler to start
	spiWaitBusy();
	return true;
}

// send _stream terminator
bool FishinoClass::spiSendTerminator(void)
{
	return spiSendSize(0xffff);
}

// check _stream terminator
bool FishinoClass::spiCheckTerminator(void)
{
	uint16_t w;
	if(!spiReceiveSize(w))
		return false;
	return w == 0xffff;
}

// send/receive following data size
bool FishinoClass::spiSendSize(uint16_t w)
{
	return spiSend((uint8_t *)&w, 2);
}

bool FishinoClass::spiReceiveSize(uint16_t &w)
{
	return spiReceive((uint8_t *)&w, 2);
}

// send a command byte
bool FishinoClass::spiSendCommand(uint8_t cmd)
{
	return spiSend(&cmd, 1);
}

// send/receive data bytes
bool FishinoClass::spiSend8(uint8_t b)
{
	if(!spiSendSize(1))
		return false;
	return spiSend(&b, 1);
}

bool FishinoClass::spiReceive8(uint8_t &b)
{
	uint16_t w;
	if(!spiReceiveSize(w) || w != 1)
		return false;
	return spiReceive(&b, 1);
}

bool FishinoClass::spiSend16(uint16_t w)
{
	if(!spiSendSize(2))
		return false;
	return spiSend((uint8_t *)&w, 2);
}

bool FishinoClass::spiReceive16(uint16_t &w)
{
	uint16_t ws;
	if(!spiReceiveSize(ws) || ws != 2)
		return false;
	return spiReceive((uint8_t *)&w, 2);
}

bool FishinoClass::spiSend32(uint32_t dw)
{
	if(!spiSendSize(4))
		return false;
	return spiSend((uint8_t *)&dw, 4);
}

bool FishinoClass::spiReceive32(uint32_t &dw)
{
	uint16_t w;
	if(!spiReceiveSize(w) || w != 4)
		return false;
	return spiReceive((uint8_t *)&dw, 4);
}

// send/receive a string
// received string is malloc-ed, MUST be freed with free()
bool FishinoClass::spiSendString(const char *str)
{
	uint16_t w = strlen(str);
	
	// first send the string length
	if(!spiSendSize(w))
		return false;

	// then send data. in chunks of 64 bytes max
	while(w)
	{
		uint16_t siz = min(64, w);
		if(!spiSend((uint8_t *)str, siz))
			return false;
		str += siz;
		w -= siz;
	}

	return true;
}

bool FishinoClass::spiReceiveString(char *&str)
{
	uint16_t w = 0;
	if(!spiReceiveSize(w))
		return false;
	
	// special case for null string
	if(w == 0x0000)
	{
		str = NULL;
		return true;
	}

	// allocate string
	str = (char *)malloc((w + 1) * sizeof(char));
	uint8_t *strP = (uint8_t *)str;
	
	// receive it in chunks of max 64 bytes
	while(w)
	{
		uint8_t siz = min(64, w);
		if(!spiReceive(strP, siz))
		{
			free(str);
			str = NULL;
			return false;
		}
		strP += siz;
		w -= siz;
	}
	*strP = 0;
	return true;
}

// send data by pointer and length
bool FishinoClass::spiSendData(byte const *buf, word bufSize)
{
	if(!spiSendSize(bufSize))
		return false;
	
	while(bufSize)
	{
		uint8_t siz = min(64, bufSize);
		if(!spiSend(buf, siz))
			return false;
		buf += siz;
		bufSize -= siz;
	}
	return true;
}

// receive data, buffer allocated, must be freed by free()
bool FishinoClass::spiReceiveDataAlloc(byte *&buf, word &bufSize)
{
	if(!spiReceiveSize(bufSize))
		return false;

	// special case for null string
	if(bufSize == 0x0000)
	{
		buf = NULL;
		return true;
	}
	
	// allocate string
	buf = (byte *)malloc((bufSize + 1) * sizeof(byte));
	
	// receive it in chunks of max 64 bytes
	byte *bufP = buf;
	uint16_t w = bufSize;
	while(w)
	{
		uint8_t siz = min(64, w);
		if(!spiReceive(bufP, w))
		{
			free(buf);
			buf = NULL;
			return false;
		}
		bufP += siz;
		w -= siz;
	}
	return true;
}

// receive data, buffer with enough space MUST be given
// bufSize MUST be initialized with buffer size
// bufSize is replaced with actual data size on success
// on error or buffer overflow, bufSize is set to 0
bool FishinoClass::spiReceiveDataFixed(uint8_t *buf, uint16_t &bufSize)
{
	uint16_t maxSize = bufSize;
	bufSize = 0;

	if(!spiReceiveSize(bufSize))
	{
		bufSize = 0;
		return false;
	}

	// if available data is greater than given buffer size
	// simply discard ALL data and return false
	if(bufSize > maxSize)
	{
		uint16_t w = bufSize;
		while(w)
		{
			uint8_t siz = min(64, w);
			if(!spiDrop(siz))
				break;
			w -= siz;
		}

		bufSize = 0;
		return false;
	}
		
	if(bufSize)
	{
		uint8_t *p = buf;
		uint16_t w = bufSize;
		while(w)
		{
			uint8_t siz = min(64, w);
			if(!spiReceive(p, siz))
			{
				bufSize = 0;
				return false;
			}
			p += siz;
			w -= siz;
		}
	}
	return true;
}

// constructor
FishinoClass::FishinoClass()
{
	// initialize firmware version
	_fwVersion = -1;
	
	// initialize current AP SSID
	_ssid = NULL;
	
	// initialize join status
	_joinStatus = STATION_IDLE;

	// initialize scanned AP info
	_apInfo = NULL;
	_numAp = 0;

	// set the slaveSelectPin as an output:
	::digitalWrite(ESP8266_CS, HIGH);
	::pinMode(ESP8266_CS, OUTPUT);
	
	// set handshake pin as input
	::pinMode(ESP8266_HANDSHAKE, INPUT);
}

// destructor
FishinoClass::~FishinoClass()
{
	// frees current ssid, if any
	if(_ssid)
		free(_ssid);
	
	// frees apInfo, if any
	freeApInfo();
}

// reset ESP and wait for it to be ready
// return true if ok, false if not ready
bool FishinoClass::reset(void)
{
	// send a status command asking for reset
	spiSelect(true);
	delay(10);

	// send a status command, value 00 (reset chip)
	SPI.transfer(0x01);		// write status
	SPI.transfer(0x00);		// 0 status code means reset
	
	delay(10);
	spiSelect(false);

	// wait till handshake pin goes low (boot mode)
	// or 5 seconds, which comes first
	uint32_t tOut = millis() + 5000;
	while(::digitalRead(ESP8266_HANDSHAKE))
		if(millis() >= tOut)
			break;
	if(millis() >= tOut)
	{
		Serial << F("RESET -- HANDSHAKE STAY HIGH\n");
		spiSelect(false);
		return false;
	}
	
	// now wait till handshake pin goes high again
	// it means that chip got initiaized after boot
	// as usual, 5 seconds timeout here
	tOut = millis() + 5000;
	while(!::digitalRead(ESP8266_HANDSHAKE))
		if(millis() >= tOut)
			break;

	// if not gone high, reset was unsuccesful
	if(!::digitalRead(ESP8266_HANDSHAKE))
	{
		Serial << F("RESET -- HANDSHAKE STAY LOW\n");
		spiSelect(false);
		return false;
	}
	
	// this one is NEEDED to give ESP8266 time to settle
	// and pre-initialize SPI interface after reset
	delay(500);

	spiSelect(true);

	// now we tell ESP to start processing commands; that's done
	// with status 1 command
	SPI.transfer(0x01);		// write status
	SPI.transfer(0x01);		// 1 status code means start main loop

	delayMicroseconds(100);
	spiSelect(false);

	delay(500);

	// and finally we shall wait for handshake pin to go low
	// so ESP is ready to receive commands from us
	tOut = millis() + 5000;
	while(::digitalRead(ESP8266_HANDSHAKE))
		if(millis() >= tOut)
			break;

	// if not gone high, reset was unsuccesful
	if(::digitalRead(ESP8266_HANDSHAKE))
	{
		Serial << F("RESET -- HANDSHAKE STAY LOW AT END\n");
		spiSelect(false);
		return false;
	}
	
	spiSelect(false);

	// give some time to settle
	delay(200);
	
	// get and check version number
	if(firmwareVersion() < FISHINO_FIRMWARE_VERSION_MIN)
	{
		Serial << F("Firmware version ") << firmwareVersionStr() << F(" is too old\r\n");
		Serial << F("Please upgrade to at least ") <<
			(FISHINO_FIRMWARE_VERSION_MIN >> 16) << F(".") <<
			((FISHINO_FIRMWARE_VERSION_MIN >> 8) & 0xff) << F(".") <<
			((FISHINO_FIRMWARE_VERSION_MIN >> 16) & 0xff)
		;
		while(true)
			;
	}
	
	return true;
}

// set spi interface parameters
bool FishinoClass::setSerialParams(uint32_t speed, uint8_t parity, uint8_t bits, uint8_t stopBits)
{
	uint8_t b;
	bool res = false;

	// activate device
	spiSelect(true);

	// send command
	if(
		spiSendCommand(CMD_SETUART)	&&
		spiSend32(speed)			&&
		spiSend8(parity)			&&
		spiSend8(bits)				&&
		spiSend8(stopBits)			&&
		spiSendTerminator()			&&
		spiReceive8(b)				&&
		spiCheckTerminator()
	)
		res = (b == UART_OK);

	spiSelect(false);
	return res;
}


// Get firmware version
uint32_t FishinoClass::firmwareVersion()
{
	
	// no need to ask twice
	if(_fwVersion != (uint32_t)-1)
		return _fwVersion;
	
	// activate device
	spiSelect(true);

	uint8_t devel, minor;
	uint16_t major;
	if(
		spiSendCommand(CMD_GETFWVERSION)	&&
		spiSendTerminator()					&&
		spiReceive8(devel)					&&
		spiReceive8(minor)					&&
		spiReceive16(major)					&&
		spiCheckTerminator()
	)
		_fwVersion = (uint32_t)major << 16 | (uint32_t)minor << 8 | devel;

	spiSelect(false);
	return _fwVersion;
}

char *FishinoClass::firmwareVersionStr()
{
	static char ver[] = "00000.000.000";
	
	uint32_t fv = firmwareVersion();
	if(fv != (uint32_t)-1)
		sprintf(ver, "%u.%u.%u", (uint16_t)((fv >> 16) & 0xffff), (uint16_t)((fv >> 8) & 0xff), (uint16_t)(fv & 0xff));
	else
		strcpy(ver, "UNKNOWN");
	return ver;
}


// set operation mode (one of WIFI_MODE)
bool FishinoClass::setMode(uint8_t mode)
{
	uint8_t b;
	bool res = false;

	// activate device
	spiSelect(true);

	if(
		spiSendCommand(CMD_SETMODE)	&&
		spiSend8(mode)				&&
		spiSendTerminator()			&&
		spiReceive8(b)				&&
		spiCheckTerminator()
	)
		res = b;

	spiSelect(false);
	return res;
}

// get operation mode
uint8_t FishinoClass::getMode(void)
{
	uint8_t b;
	uint8_t mode = NULL_MODE;

	// activate device
	spiSelect(true);

	if(
		spiSendCommand(CMD_QUERYMODE)	&&
		spiSendTerminator()				&&
		spiReceive8( b)					&&
		spiCheckTerminator()
	)
		mode = b;

	spiSelect(false);
	return mode;
}

// Start Wifi connection for OPEN networks
// param ssid: Pointer to the SSID string.
// return one of JOIN_STATUS result
uint8_t FishinoClass::begin(const char* ssid)
{
	return begin(ssid, "");
}

// Start Wifi connection with encryption.
// return one of JOIN_STATUS result
uint8_t FishinoClass::begin(const char* ssid, const char *passphrase)
{
	// freea any previous SSID
	if(_ssid)
	{
		free(_ssid);
		_ssid = NULL;
	}
	_joinStatus = STATION_TIMEOUT;
	
	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		spiSendCommand(CMD_JOINAP)	&&
		spiSendString(ssid)			&&
		spiSendString(passphrase)	&&
		spiSendTerminator()
	)
	{
		// try to join for 20 seconds maximum
		uint8_t b = STATION_TIMEOUT;
		if(spiWaitDataAvailable(20000))
		{
			// get status byte back
			if(
				spiReceive8(b)			&&
				spiCheckTerminator()
			)
			{
				if(b == STATION_GOT_IP)
					_ssid = strdup(ssid);
				_joinStatus = b;
			}
		}
	}
	
	spiSelect(false);
	return _joinStatus;
}

// Change Ip configuration settings disabling the dhcp client
//	param local_ip: 	Static ip configuration
bool FishinoClass::config(IPAddress local_ip)
{
	uint8_t b;
	bool res = false;
	
	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		spiSendCommand(CMD_SETSTAIP)	&&
		spiSendData(&local_ip[0], 4)	&&
		spiSendTerminator()				&&
		spiReceive8(b)					&&
		spiCheckTerminator()
	)
		res = (b == IP_OK);

	spiSelect(false);
	return res;
}

// Change Ip configuration settings disabling the dhcp client
// param local_ip:			Static ip configuration
// param dns_server:		IP configuration for DNS server
// DNS SERVER SETTING NOT SUPPORTED BY NOW -- DUMMY
bool FishinoClass::config(IPAddress local_ip, IPAddress dns_server)
{
	return config(local_ip);
}

// Change Ip configuration settings disabling the dhcp client
// param local_ip:			Static ip configuration
// param dns_server:		IP configuration for DNS server 1
// param gateway :			Static gateway configuration
// DNS SERVER SETTING NOT SUPPORTED BY NOW -- DUMMY
bool FishinoClass::config(IPAddress local_ip, IPAddress dns_server, IPAddress gateway)
{
	uint8_t b;
	bool res = false;

	if(!config(local_ip))
		return false;

	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		spiSendCommand(CMD_SETSTAGW)	&&
		spiSendData(&gateway[0], 4)		&&
		spiSendTerminator()				&&
		spiReceive8(b)					&&
		spiCheckTerminator()
	)
		res = (b == IP_OK);

	spiSelect(false);
	return res;
}

// Change Ip configuration settings disabling the dhcp client
// param local_ip:			Static ip configuration
// param dns_server:		IP configuration for DNS server 1
// param gateway:			Static gateway configuration
// param subnet:			Static Subnet mask
// DNS SERVER SETTING NOT SUPPORTED BY NOW -- DUMMY
bool FishinoClass::config(IPAddress local_ip, IPAddress dns_server, IPAddress gateway, IPAddress subnet)
{
	uint8_t b;
	bool res = false;

	if(!config(local_ip, IPAddress()))
		return false;

	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		spiSendCommand(CMD_SETSTANM)	&&
		spiSendData(&subnet[0], 4)		&&
		spiSendTerminator()				&&
		spiReceive8(b)					&&
		spiCheckTerminator()
	)
		res = (b == IP_OK);

	spiSelect(false);
	return res;
}

// Change DNS Ip configuration
// param dns_server1:		ip configuration for DNS server 1
// DNS SERVER SETTING NOT SUPPORTED BY NOW -- DUMMY
bool FishinoClass::setDNS(IPAddress dns_server1)
{
	return true;
}

// Change DNS Ip configuration
// param dns_server1:		ip configuration for DNS server 1
// param dns_server2:		ip configuration for DNS server 2
// DNS SERVER SETTING NOT SUPPORTED BY NOW -- DUMMY
bool FishinoClass::setDNS(IPAddress dns_server1, IPAddress dns_server2)
{
	return true;
}

// Disconnect from the network
// return: one value of wl_status_t enum
bool FishinoClass::disconnect(void)
{
	uint8_t b;
	bool res = false;

	// activate device
	spiSelect(true);

	if(
		spiSendCommand( CMD_QUITAP)	&&
		spiSendTerminator()			&&
		spiReceive8(b)				&&
		spiCheckTerminator()
	)
		res = true;

	spiSelect(false);
	return res;
}

// Get the interface MAC address.
// return: pointer to A STATIC uint8_t array with length WL_MAC_ADDR_LENGTH
const uint8_t* FishinoClass::macAddress(void)
{
	static uint8_t MAC[6];
	uint16_t macBufLen = sizeof(MAC);

	// activate device
	spiSelect(true);

	if(
		!spiSendCommand( CMD_GETSTAMAC)			||
		!spiSendTerminator()					||
		!spiReceiveDataFixed(MAC, macBufLen)	||
		!spiCheckTerminator()
	)
		memset(MAC, macBufLen, 0);
	
	spiSelect(false);
	return MAC;
}

// Get the interface IP address.
// return: Ip address value
IPAddress FishinoClass::localIP()
{
	uint8_t buf[4];
	uint16_t bufLen = sizeof(buf);

	// activate device
	spiSelect(true);

	if(
		!spiSendCommand( CMD_GETSTAIP)		||
		!spiSendTerminator()				||
		!spiReceiveDataFixed(buf, bufLen)	||
		!spiCheckTerminator()
	)
		memset(buf, bufLen, 0);

	spiSelect(false);
	return IPAddress(buf);
}

// Get the interface subnet mask address.
// return: subnet mask address value
IPAddress FishinoClass::subnetMask()
{
	IPAddress ip;
	uint16_t ipBufLen = sizeof(IPAddress);

	// activate device
	spiSelect(true);

	if(
		!spiSendCommand( CMD_GETSTANM)					||
		!spiSendTerminator()							||
		!spiReceiveDataFixed((uint8_t *)&ip, ipBufLen)	||
		!spiCheckTerminator()
	)
		ip = IPAddress();
	
	spiSelect(false);
	return ip;
}

// Get the gateway ip address.
// return: gateway ip address value
IPAddress FishinoClass::gatewayIP()
{
	IPAddress ip;
	uint16_t ipBufLen = sizeof(IPAddress);

	// activate device
	spiSelect(true);

	if(
		!spiSendCommand(CMD_GETSTAGW)					||
		!spiSendTerminator()							||
		!spiReceiveDataFixed((uint8_t *)&ip, ipBufLen)	||
		!spiCheckTerminator()
	)
		ip = IPAddress();

	spiSelect(false);
	return ip;
}

// Return the current SSID associated with the network
// return: ssid string
const char* FishinoClass::SSID()
{
	return _ssid;
}

// Return the current BSSID associated with the network.
// It is the MAC address of the Access Point
// return: pointer to uint8_t array with length WL_MAC_ADDR_LENGTH
const uint8_t* FishinoClass::BSSID()
{
	static const uint8_t dummy[6] = {0, 0, 0, 0, 0, 0};
	
	if(!_ssid)
		return dummy;
	APINFO const *info = findApInfo(_ssid);
	if(!info)
	{
		scanNetworks();
		info = findApInfo(_ssid);
	}
	if(!info)
		return dummy;
	return info->bssid;
}

// Return the current RSSI /Received Signal Strength in dBm)
// associated with the network
// return: signed value
int32_t FishinoClass::RSSI()
{
	if(!_ssid)
		return 0;
	APINFO const *info = findApInfo(_ssid);
	if(!info)
	{
		scanNetworks();
		info = findApInfo(_ssid);
	}
	if(!info)
		return 0;
	return info->rssi;
}

// Return the Encryption Type associated with the network
// return: one value of wl_enc_type enum
uint8_t	FishinoClass::encryptionType()
{
	if(!_ssid)
		return 0;
	APINFO const *info = findApInfo(_ssid);
	if(!info)
	{
		scanNetworks();
		info = findApInfo(_ssid);
	}
	if(!info)
		return 0;
	return info->authmode;
}

// Start scan WiFi networks available
// return: Number of discovered networks
uint8_t FishinoClass::scanNetworks()
{
	// frees any previous ap info
	freeApInfo();

	// activate device
	spiSelect(true);

	if(
		!spiSendCommand(CMD_LISTAP)	||
		!spiSendTerminator()
	)
	{
		spiSelect(false);
		return 0;
	}
		
	// try 10 seconds to get ap list
	byte b;
	if(
		!spiWaitDataAvailable(10000) ||
		!spiReceive8(b)
	)
	{
		spiSelect(false);
		return 0;
	}

	if(b != AP_SCAN_DONE)
	{
		spiCheckTerminator();
		spiSelect(false);
		return 0;
	}

	// get number of aps
	if(!spiReceive8(_numAp))
	{
		_numAp = 0;
		spiSelect(false);
		return 0;
	}

	// allocate APINFO array
	_apInfo = (APINFO *)malloc(_numAp * sizeof(APINFO));
	memset(_apInfo, 0, _numAp * sizeof(APINFO));
	
	for(int i = 0; i < _numAp; i++)
	{
		// read authmode
		uint8_t authMode;
		char *cSSID = NULL;
		char RSSI;
		uint8_t mac[6];
		uint16_t macLen = 6;
		uint8_t channel;

		if(
			!spiReceive8(authMode)				||
			!spiReceiveString(cSSID)			||
			!spiReceive8((byte &)RSSI)			||
			!spiReceiveDataFixed(mac, macLen)	||
			!spiReceive8(channel)
		)
		{
			freeApInfo();
			spiSelect(false);
			return 0;
		}
		APINFO *curInfo = &_apInfo[i];
		curInfo->authmode = authMode;
		curInfo->ssid = cSSID;
		curInfo->rssi = RSSI;
		curInfo->channel = channel;
		memcpy(curInfo->bssid, mac, 6);
	}

	spiCheckTerminator();

	spiSelect(false);
	
	return _numAp;
}

// Return the SSID discovered during the network scan.
// param networkItem: specify from which network item want to get the information
// return: ssid string of the specified item on the networks scanned list
const char* FishinoClass::SSID(uint8_t networkItem)
{
	if(networkItem >= _numAp)
		return NULL;
	return _apInfo[networkItem].ssid;
}

// Return the encryption type of the networks discovered during the scanNetworks
// param networkItem: specify from which network item want to get the information
// return: encryption type (enum wl_enc_type) of the specified item on the networks scanned list
uint8_t FishinoClass::encryptionType(uint8_t networkItem)
{
	if(networkItem >= _numAp)
		return 0;
	return _apInfo[networkItem].authmode;
}

// Return the RSSI of the networks discovered during the scanNetworks
// param networkItem: specify from which network item want to get the information
// return: signed value of RSSI of the specified item on the networks scanned list
int32_t FishinoClass::RSSI(uint8_t networkItem)
{
	if(networkItem >= _numAp)
		return 0;
	return _apInfo[networkItem].rssi;
}

// Return Connection status.
// return: one of the value defined in JOIN_STATUS
uint8_t FishinoClass::status()
{
	// activate device
	spiSelect(true);

	if(
		!spiSendCommand(CMD_QUERYAP)	||
		!spiSendTerminator()
	)
	{
		spiSelect(false);
		return STATION_IDLE;
	}
		
	// try 1 second max to get response
	byte b;
	if(
		!spiWaitDataAvailable(1000) ||
		!spiReceive8(b)
	)
	{
		spiSelect(false);
		return STATION_IDLE;
	}
	
	_joinStatus = b;
	
	// check if station is connected
	if(b != STATION_GOT_IP)
	{
		spiCheckTerminator();
		spiSelect(false);
		return STATION_IDLE;
	}
	
	// if connected, gather the SSID
	char *cSSID = NULL;
	if(
		!spiReceiveString(cSSID)	||
		!spiCheckTerminator()
	)
	{
		if(cSSID)
			free(cSSID);
		spiSelect(false);
		return STATION_IDLE;
	}
	if(_ssid)
		free(_ssid);
	_ssid = cSSID;

	spiSelect(false);
	return STATION_GOT_IP;
}

// Resolve the given hostname to an IP address.
// param aHostname: Name to be resolved
// param aResult: IPAddress structure to store the returned IP address
// result: 1 if aIPAddrString was successfully converted to an IP address,
// else error code
bool FishinoClass::hostByName(const char* aHostname, IPAddress& aResult)
{
	// STILL NOT IMPLEMENTED
	return false;
}

// low level TCP stuffs -- used by client and server

// connect to host -- return socket number, or 0xff if error
uint8_t FishinoClass::connect(const char *host, uint16_t port)
{
	// activate device
	spiSelect(true);

	if(
		!spiSendCommand(CMD_CONNECT)	||
		// type 0 == TCP
		!spiSend8(0)					||
		!spiSendString(host)			||
		!spiSend32(port)				||
		!spiSendTerminator()
	)
	{
		spiSelect(false);
		return 0xff;
	}

	// try 5 seconds to connect
	uint16_t res = 0;
	if(
		!spiWaitDataAvailable(5000)	||
		!spiReceive16(res)			||
		!spiCheckTerminator()
	)
	{
		spiSelect(false);
		return 0xff;
	}

	if(res > 255)
		res = 255;

	spiSelect(false);
	return res;
}

// connect to a secure host (HTTPS - SSL)
uint8_t FishinoClass::secureConnect(const char *host, uint16_t port)
{
	// activate device
	spiSelect(true);

	if(
		!spiSendCommand(CMD_CONNECT)	||
		// type 2 == TCP+SSL
		!spiSend8(2)					||
		!spiSendString(host)			||
		!spiSend32(port)				||
		!spiSendTerminator()
	)
	{
		spiSelect(false);
		return 0xff;
	}

	// try 5 seconds to connect
	uint16_t res = 0;
	if(
		!spiWaitDataAvailable(5000)	||
		!spiReceive16(res)			||
		!spiCheckTerminator()
	)
	{
		spiSelect(false);
		return 0xff;
	}

	if(res > 255)
		res = 255;

	spiSelect(false);
	return res;
}


// disconnect from host -- return true on success, false otherwise
bool FishinoClass::disconnect(uint8_t sock)
{
	// activate device
	spiSelect(true);

	if(
		!spiSendCommand(CMD_CLOSE)	||
		!spiSend8(sock)				||
		!spiSendTerminator()
	)
	{
		spiSelect(false);
		return false;
	}
	
	// try 5 seconds to disconnect
	byte res;
	if(
		!spiWaitDataAvailable(5000)		||
		!spiReceive8(res)				||
		!spiCheckTerminator()
	)
	{
		spiSelect(false);
		return false;
	}

	spiSelect(false);
	return res == DISCONNECTION_DISCONNECTED;
}

// write to socket -- return true on success, false otherwise
bool FishinoClass::write(uint8_t sock, uint8_t const *buf, uint16_t bufLen)
{
	// activate device
	spiSelect(true);

/*
unsigned long tim0 = micros();
*/
	if(
		!spiSendCommand(CMD_SEND)	||
		!spiSend8(sock)				||
		!spiSendData(buf, bufLen)	||
		!spiSendTerminator()
	)
	{
		spiSelect(false);
		return false;
	}
/*
unsigned long tim1 = micros();
*/

	if(!spiWaitDataAvailable(5000))
	{
		spiSelect(false);
		return false;
	}
/*
unsigned long tim2 = micros();
*/
	uint8_t res;
	if(
		!spiReceive8(res)		||
		!spiCheckTerminator()
	)
	{
		spiSelect(false);
		return false;
	}
/*
unsigned long tim3 = micros();
tim3 -= tim2;
tim2 -= tim1;
tim1 -= tim0;
Serial << F("Timing bytes = ") << bufLen << F(" T1 = ") << tim1 << F(" T2 = ") << tim2 << F(" T3 = ") << tim3 << "\n";
*/
	spiSelect(false);
	return res == SEND_OK;
}

// read from socket -- return true on success, false otherwise
// buffer MUST be big enough to contain reqBytes
// on exit, bufLen contains number of bytes actually read
bool FishinoClass::read(uint8_t sock, uint16_t reqBytes, uint8_t *buf, uint16_t &bufLen)
{
	// activate device
	spiSelect(true);

	if(
		!spiSendCommand(CMD_READ)	||
		!spiSend8(sock)				||
		!spiSend16(reqBytes)		||
		!spiSendTerminator()
	)
	{
		spiSelect(false);
		return false;
	}

	// try 5 seconds to get data
	if(!spiWaitDataAvailable(5000))
	{
		spiSelect(false);
		return false;
	}
	byte stat;
	if(!spiReceive8(stat))
	{
		spiSelect(false);
		return false;
	}
	
	// we espect receive ok status
	if(stat != RECEIVE_OK)
	{
		spiCheckTerminator();
		spiSelect(false);
		return false;
	}
	
	bufLen = reqBytes;
	if(!spiReceiveDataFixed(buf, bufLen))
	{
		spiSelect(false);
		return false;
	}

	spiCheckTerminator();
	spiSelect(false);
	return true;
}

// return number of available bytes on socket
uint16_t FishinoClass::available(uint8_t sock)
{
	// activate device
	spiSelect(true);

	if(
		!spiSendCommand(CMD_AVAIL)	||
		!spiSend8(sock)				||
		!spiSendTerminator()
	)
	{
		spiSelect(false);
		return 0;
	}
	
	// try 5 seconds to get data
	byte stat;
	if(
		!spiWaitDataAvailable(5000)		||
		!spiReceive8(stat)
	)
	{
		spiSelect(false);
		return 0;
	}
	
	// we espect receive ok status
	if(stat != RECEIVE_OK)
	{
		spiCheckTerminator();
		spiSelect(false);
		return 0;
	}
	
	// status ok, retrieve number of bytes
	uint16_t size;
	if(!spiReceive16(size))
	{
		spiSelect(false);
		return 0;
	}

	spiCheckTerminator();
	spiSelect(false);
	return size;
}

// peek a data byte from socket, if any
// returns true on success, false otherwise
bool FishinoClass::peek(uint8_t sock, uint8_t &b)
{
	// activate device
	spiSelect(true);

	if(
		!spiSendCommand(CMD_PEEK)	||
		!spiSend8(sock)				||
		!spiSendTerminator()
	)
	{
		spiSelect(false);
		return false;
	}
	
	// try 5 seconds to get data
	byte stat;
	if(
		!spiWaitDataAvailable(5000)		||
		!spiReceive8(stat)
	)
	{
		spiSelect(false);
		return false;
	}
	
	// we espect receive ok status
	if(stat != RECEIVE_OK)
	{
		spiCheckTerminator();
		spiSelect(false);
		return false;
	}
	
	// get peeked byte
	if(
		!spiReceive8(b)	||
		!spiCheckTerminator()
	)
	{
		spiSelect(false);
		return false;
	}
	spiSelect(false);
	return true;
}

// flush socket data
bool FishinoClass::flush(uint8_t sock)
{
	// activate device
	spiSelect(true);

	if(
		!spiSendCommand(CMD_FLUSH)	||
		!spiSend8(sock)				||
		!spiSendTerminator()
	)
	{
		spiSelect(false);
		return false;
	}
	
	// try 5 seconds to get data
	byte stat;
	if(
		!spiWaitDataAvailable(5000)		||
		!spiReceive8(stat)				||
		!spiCheckTerminator()
	)
	{
		spiSelect(false);
		return false;
	}
	
	spiSelect(false);
	return stat == RECEIVE_OK;
}

// connection status
uint8_t FishinoClass::status(uint8_t sock)
{
	// activate device
	spiSelect(true);

	if(
		!spiSendCommand(CMD_CONNSTATUS)	||
		!spiSend8(sock)					||
		!spiSendTerminator()
	)
	{
		spiSelect(false);
		return SOCK_INVALID;
	}
	
	// try 5 seconds to get data
	byte stat;
	if(
		!spiWaitDataAvailable(5000)		||
		!spiReceive8(stat)				||
		!spiCheckTerminator()
	)
	{
		spiSelect(false);
		return SOCK_INVALID;
	}
	
	spiSelect(false);
	return stat;
}

// start a server on given port
// ONLY ONE SERVER AT A TIME
bool FishinoClass::startServer(uint32_t port, uint16_t timeout)
{
	// activate device
	spiSelect(true);

	if(
		!spiSendCommand(CMD_STARTSERVER)	||
		!spiSend32(port)					||
		!spiSend16(timeout)					||
		!spiSendTerminator()
	)
	{
		spiSelect(false);
		return false;
	}

	// try 5 seconds to get data
	byte stat;
	if(
		!spiWaitDataAvailable(5000)		||
		!spiReceive8(stat)				||
		!spiCheckTerminator()
	)
	{
		spiSelect(false);
		return TCPSERVER_UART_TIMEOUT;
	}

	spiSelect(false);
	return (stat == TCPSERVER_LISTENING || stat == TCPSERVER_ALREADY);
}

// stops server
bool FishinoClass::stopServer(void)
{
	// activate device
	spiSelect(true);

	if(
		!spiSendCommand(CMD_STOPSERVER)		||
		!spiSendTerminator()
	)
	{
		spiSelect(false);
		return false;
	}

	// try 5 seconds to get data
	byte stat;
	if(
		!spiWaitDataAvailable(5000)		||
		!spiReceive8(stat)				||
		!spiCheckTerminator()
	)
	{
		spiSelect(false);
		return false;
	}

	spiSelect(false);
	return stat == TCPSERVER_OFF;
}

// poll server for sockets with data
// returns a dynamic allocated buffer which MUST be freed by free
bool FishinoClass::pollServer(bool all, uint8_t *&buf, uint16_t &count)
{
	// activate device
	spiSelect(true);

	if(
		!spiSendCommand(CMD_POLLSERVER)	||
		!spiSend8(all)					||
		!spiSendTerminator()
	)
	{
		spiSelect(false);
		return false;
	}

	// try 5 seconds to get data
	if(!spiWaitDataAvailable(5000))
	{
		spiSelect(false);
		return false;
	}

	buf = NULL;
	count = 0;
	if(
		!spiReceiveDataAlloc(buf, count)	||
		!spiCheckTerminator()
	)
	{
		if(buf)
			free(buf);
		buf = NULL;
		count = 0;
		spiSelect(false);
		return false;
	}
	spiSelect(false);
	return true;
}

// "uniform" functions to set station and AP parameters
// station ones replicates the above ones, jus to have meaningful names
bool FishinoClass::setStaIP(IPAddress ip)
{
	return config(ip);
}

bool FishinoClass::setStaMAC(uint8_t const *mac)
{
	uint8_t b;
	bool res = false;
	
	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		spiSendCommand(CMD_SETSTAMAC)	&&
		spiSendData(mac, 6)				&&
		spiSendTerminator()				&&
		spiReceive8(b)					&&
		spiCheckTerminator()
	)
		res = (b == MAC_OK);

	spiSelect(false);
	return res;
}

bool FishinoClass::setStaGateway(IPAddress gw)
{
	uint8_t b;
	bool res = false;
	
	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		spiSendCommand(CMD_SETSTAGW)	&&
		spiSendData(&gw[0], 4)			&&
		spiSendTerminator()				&&
		spiReceive8(b)					&&
		spiCheckTerminator()
	)
		res = (b == IP_OK);

	spiSelect(false);
	return res;
}

bool FishinoClass::setStaNetMask(IPAddress nm)
{
	uint8_t b;
	bool res = false;
	
	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		spiSendCommand(CMD_SETSTANM)	&&
		spiSendData(&nm[0], 4)			&&
		spiSendTerminator()				&&
		spiReceive8(b)					&&
		spiCheckTerminator()
	)
		res = (b == IP_OK);

	spiSelect(false);
	return res;
}

bool FishinoClass::staStartDHCP(void)
{
	uint8_t b;
	bool res = false;
	
	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		spiSendCommand(CMD_ENABLESTADHCP)	&&
		spiSendTerminator()					&&
		spiReceive8(b)						&&
		spiCheckTerminator()
	)
		res = (b == DHCP_OK);

	spiSelect(false);
	return res;
}

bool FishinoClass::staStopDHCP(void)
{
	uint8_t b;
	bool res = false;
	
	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		spiSendCommand(CMD_DISABLESTADHCP)	&&
		spiSendTerminator()					&&
		spiReceive8(b)						&&
		spiCheckTerminator()
	)
		res = (b == DHCP_OK);

	spiSelect(false);
	return res;
}

bool FishinoClass::getStaDHCPStatus(void)
{
	uint8_t b;
	bool res = false;
	
	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		spiSendCommand(CMD_QUERYSTADHCP)	&&
		spiSendTerminator()					&&
		spiReceive8(b)						&&
		spiCheckTerminator()
	)
		res = b;

	spiSelect(false);
	return res;
}

bool FishinoClass::setApIP(IPAddress ip)
{
	uint8_t b;
	bool res = false;
	
	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		spiSendCommand(CMD_SETAPIP)		&&
		spiSendData(&ip[0], 4)	&&
		spiSendTerminator()				&&
		spiReceive8(b)					&&
		spiCheckTerminator()
	)
		res = (b == IP_OK);

	spiSelect(false);
	return res;
}

bool FishinoClass::setApMAC(uint8_t const *mac)
{
	uint8_t b;
	bool res = false;
	
	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		spiSendCommand(CMD_SETAPMAC)	&&
		spiSendData(mac, 6)				&&
		spiSendTerminator()				&&
		spiReceive8(b)					&&
		spiCheckTerminator()
	)
		res = (b == MAC_OK);

	spiSelect(false);
	return res;
}

bool FishinoClass::setApGateway(IPAddress gw)
{
	uint8_t b;
	bool res = false;
	
	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		spiSendCommand(CMD_SETAPGW)		&&
		spiSendData(&gw[0], 4)			&&
		spiSendTerminator()				&&
		spiReceive8(b)					&&
		spiCheckTerminator()
	)
		res = (b == IP_OK);

	spiSelect(false);
	return res;
}

bool FishinoClass::setApNetMask(IPAddress nm)
{
	uint8_t b;
	bool res = false;
	
	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		spiSendCommand(CMD_SETAPNM)		&&
		spiSendData(&nm[0], 4)			&&
		spiSendTerminator()				&&
		spiReceive8(b)					&&
		spiCheckTerminator()
	)
		res = (b == IP_OK);

	spiSelect(false);
	return res;
}

// full AP IP info
bool FishinoClass::setApIPInfo(IPAddress ip, IPAddress gateway, IPAddress netmask)
{
	uint8_t b;
	bool res = false;
	
	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		spiSendCommand(CMD_SETAPIPINFO)	&&
		spiSendData(&ip[0], 4)			&&
		spiSendData(&gateway[0], 4)		&&
		spiSendData(&netmask[0], 4)		&&
		spiSendTerminator()				&&
		spiReceive8(b)					&&
		spiCheckTerminator()
	)
		res = (b == IP_OK);

	spiSelect(false);
	return res;
}

// softAp DHCP
bool FishinoClass::softApStartDHCPServer(void)
{
	uint8_t b;
	bool res = false;
	
	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		spiSendCommand(CMD_ENABLEAPDHCP)	&&
		spiSendTerminator()					&&
		spiReceive8(b)						&&
		spiCheckTerminator()
	)
		res = (b == DHCP_OK);

	spiSelect(false);
	return res;
}

bool FishinoClass::softApStartDHCPServer(IPAddress startIP, IPAddress endIP)
{
	uint8_t b;
	bool res = false;
	
	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		spiSendCommand(CMD_ENABLEAPDHCP)	&&
		spiSendData(&startIP[0], 4)			&&
		spiSendData(&endIP[0], 4)			&&
		spiSendTerminator()					&&
		spiReceive8(b)						&&
		spiCheckTerminator()
	)
		res = (b == DHCP_OK);

	spiSelect(false);
	return res;
}

bool FishinoClass::softApStopDHCPServer(void)
{
	uint8_t b;
	bool res = false;
	
	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		spiSendCommand(CMD_DISABLEAPDHCP)	&&
		spiSendTerminator()					&&
		spiReceive8(b)						&&
		spiCheckTerminator()
	)
		res = (b == DHCP_OK);

	spiSelect(false);
	return res;
}

bool FishinoClass::getSoftApDHCPServerStatus(void)
{
	uint8_t b;
	bool res = false;
	
	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		spiSendCommand(CMD_QUERYAPDHCPS)	&&
		spiSendTerminator()					&&
		spiReceive8(b)						&&
		spiCheckTerminator()
	)
		res = b;

	spiSelect(false);
	return res;
}

// softAp configuration

// get ap SSID - return a dynamic buffer that MUST be freed by free()
// return NULL on error
char *FishinoClass::softApGetSSID(void)
{
	char *SSID, *pass;
	uint8_t channel;
	bool hidden;
	if(!softApGetConfig(SSID, pass, channel, hidden))
		return NULL;
	if(pass)
		free(pass);
	return SSID;
}

// get ap PASSWORD - return a dynamic buffer that MUST be freed by free()
// return NULL on error
char *FishinoClass::softApGetPassword(void)
{
	char *SSID, *pass;
	uint8_t channel;
	bool hidden;
	if(!softApGetConfig(SSID, pass, channel, hidden))
		return NULL;
	if(SSID)
		free(SSID);
	return pass;
}

// return ap channel
// return 0 on error
uint8_t FishinoClass::softApGetChannel(void)
{
	char *SSID, *pass;
	uint8_t channel;
	bool hidden;
	if(!softApGetConfig(SSID, pass, channel, hidden))
		return 0;
	if(SSID)
		free(SSID);
	if(pass)
		free(pass);
	return channel;
}

// return ap hidden state
bool FishinoClass::softApGetHidden(void)
{
	char *SSID, *pass;
	uint8_t channel;
	bool hidden;
	if(!softApGetConfig(SSID, pass, channel, hidden))
		return false;
	if(SSID)
		free(SSID);
	if(pass)
		free(pass);
	return hidden;
}

// get all softAp config data
// warning - all returned strings MUST be freed by free()
bool FishinoClass::softApGetConfig(char *&SSID, char *&pass, uint8_t &channel, bool &hidden)
{
	// activate device
	spiSelect(true);

	byte b;
	if(
		!spiSendCommand(CMD_APGETCONFIG)	||
		!spiSendTerminator()				||
		!spiReceive8( b)
	)
	{
		spiSelect(false);
		return false;
	}

	// check result status
	if(b != AP_CONFIG_OK)
	{
		spiCheckTerminator();
		spiSelect(false);
		return false;
	}

	bool err = false;
	SSID = NULL;
	pass = NULL;

	// get SSID
	if(!spiReceiveString(SSID))
		err = true;
	
	// get password
	if(!err && !spiReceiveString(pass))
		err = true;
	
	// get channel
	if(!err && !spiReceive8(channel))
		err = true;
	
	// get hidden state
	byte hid;
	if(!err && !spiReceive8(hid))
		err = true;
	hidden = (hid ? true : false);
	
	// check terminator
	if(!err && !spiCheckTerminator())
		err = true;
	
	spiSelect(false);
	
	// on error, frees strings
	if(err)
	{
		if(SSID)
			free(SSID);
		SSID = NULL;
		if(pass)
			free(pass);
		pass = NULL;
		return false;
	}
	
	return true;
}

// set softAp parameters
bool FishinoClass::softApConfig(const char *SSID, const char *pass, uint8_t channel, bool hidden)
{
	// activate device
	spiSelect(true);

	if(
		!spiSendCommand(CMD_APSETCONFIG)	||
		!spiSendString(SSID)				||
		!spiSendString(pass)				||
		!spiSend8(channel)					||
		!spiSend8(hidden)					||
		!spiSendTerminator()
	)
	{
		spiSelect(false);
		return false;
	}
	
	// get response back
	byte b;
	if(
		!spiReceive8(b)		||
		!spiCheckTerminator()
	)
	{
		spiSelect(false);
		return false;
	}
	spiSelect(false);
	return b == AP_CONFIG_OK;
}

// setup station AP parameters without joining
bool FishinoClass::setStaConfig(const char *ssid)
{
	return setStaConfig(ssid, "");
}

bool FishinoClass::setStaConfig(const char *ssid, const char *passphrase)
{
	uint8_t res = 0;
	
	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		!spiSendCommand(CMD_SETSTACONFIG)	||
		!spiSendString(ssid)				||
		!spiSendString(passphrase)			||
		!spiSendTerminator()				||
		!spiReceive8(res)					||
		!spiCheckTerminator()
	)
		res = 0;
	
	spiSelect(false);
	return res;
}

// get current station AP parameters
// WARNING : returns dynamic buffers wich MUST be freed
bool FishinoClass::getStaConfig(char *&ssid, char *&pass)
{
	ssid = NULL;
	pass = NULL;
	uint8_t *mac = NULL;
	uint16_t macSize = 0;
	uint8_t res = 0;

	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		!spiSendCommand(CMD_GETSTACONFIG)	||
		!spiSendTerminator()				||
		!spiReceive8(res)
	)
		res = 0;
	if(!res)
		spiCheckTerminator();
	else
	{
		if(
			!spiReceiveString(ssid)				||
			!spiReceiveString(pass)				||
			!spiReceiveDataAlloc(mac, macSize)	||
			!spiCheckTerminator()
		)
		{
			if(ssid)
				free(ssid);
			if(pass)
				free(pass);
			if(mac)
				free(mac);
			ssid = NULL;
			pass = NULL;
			mac = NULL;
			res = 0;
		}
	}
	
	// we don't need mac
	if(mac)
		free(mac);

	spiSelect(false);
	return res;
	
}

// join AP setup by setStaConfig call (or previous AP set by begin)
uint8_t FishinoClass::joinAp(void)
{
	// free any previous SSID
	if(_ssid)
	{
		free(_ssid);
		_ssid = NULL;
	}
	_joinStatus = STATION_TIMEOUT;
	
	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		spiSendCommand(CMD_STACONNECT)	&&
		spiSendTerminator()
	)
	{
		// try to join for 20 seconds maximum
		uint8_t b = STATION_TIMEOUT;
		if(spiWaitDataAvailable(20000))
		{
			// get status byte back
			if(
				spiReceive8(b)			&&
				spiCheckTerminator()
			)
				_joinStatus = b;
		}
	}
	
	spiSelect(false);

	if(_joinStatus == STATION_GOT_IP)
	{
		// get ssid from esp
		char *pass;
		if(getStaConfig(_ssid, pass))
			free(pass);
	}

	return _joinStatus;
}

// quits AP
bool FishinoClass::quitAp(void)
{
	uint8_t res = 0;
	
	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		!spiSendCommand(CMD_STADISCONNECT)	||
		!spiSendTerminator()				||
		!spiReceive8(res)					||
		!spiCheckTerminator()
	)
		res = 0;

	spiSelect(false);
	return res;
}

////////////////////////////////////////////////////////////////////////
// extra I/O pins on ESP module

// pinMode for ESP I/O
bool FishinoClass::pinMode(uint8_t pin, uint8_t mode)
{
	uint8_t res = 0;
	
	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		!spiSendCommand(CMD_PINMODE)	||
		!spiSend8(pin)					||
		!spiSend8(mode)					||
		!spiSendTerminator()			||
		!spiReceive8(res)				||
		!spiCheckTerminator()
	)
		res = 0;

	spiSelect(false);
	return res;
}

// digital I/O for ESP I/O
uint8_t FishinoClass::digitalRead(uint8_t pin)
{
	uint8_t res = 0;
	uint8_t val;
	
	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		!spiSendCommand(CMD_DIGITALREAD)	||
		!spiSend8(pin)						||
		!spiSendTerminator()				||
		!spiReceive8(res)
	)
		res = 0;
	if(res)
	{
		if(!spiReceive8(val))
			res = 0;
	}
	if(!res)
		val = 0;

	spiCheckTerminator();
	spiSelect(false);
	return val;
}

bool FishinoClass::digitalWrite(uint8_t pin, uint8_t val)
{
	uint8_t res = 0;
	
	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		!spiSendCommand(CMD_DIGITALWRITE)	||
		!spiSend8(pin)						||
		!spiSend8(val)						||
		!spiSendTerminator()				||
		!spiReceive8(res)					||
		!spiCheckTerminator()
	)
		res = 0;

	spiSelect(false);
	return res;
}

// analog read
uint16_t FishinoClass::analogRead(void)
{
	uint8_t res = 0;
	uint16_t val;
	
	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		!spiSendCommand(CMD_ANALOGREAD)		||
		!spiSendTerminator()				||
		!spiReceive8(res)
	)
		res = 0;
	if(res)
	{
		if(!spiReceive16(val))
			res = 0;
	}
	if(!res)
		val = 0;

	spiCheckTerminator();
	spiSelect(false);
	return val;
}

// physical mode handling
uint8_t FishinoClass::getPhyMode(void)
{
	uint8_t res = 0;
	uint8_t mode;
	
	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		!spiSendCommand(CMD_GETPHYMODE)	||
		!spiSendTerminator()			||
		!spiReceive8(res)
	)
		res = 0;
	if(res)
	{
		if(!spiReceive8(mode))
			res = 0;
	}
	if(!res)
		mode = 0;

	spiCheckTerminator();
	spiSelect(false);
	return mode;
}

bool FishinoClass::setPhyMode(uint8_t mode)
{
	uint8_t res = 0;
	
	// activate device
	spiSelect(true);

	// send request to ESP
	if(
		!spiSendCommand(CMD_SETPHYMODE)	||
		!spiSend8(mode)					||
		!spiSendTerminator()			||
		!spiReceive8(res)				||
		!spiCheckTerminator()
	)
		res = 0;

	spiSelect(false);
	return res;
}

FishinoClass Fishino;
