/*
  Fishino.h - Library for ESP8266 WiFi module
  Copyright (c) 2015 MicioMax.  All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef FISHINO_H
#define FISHINO_H

// macro to build 32 bit version number from components
#define VERSION(maj, min, dev) ( ((uint32_t)maj << 16) | ((uint32_t)(min & 0xFF)) << 8 | ((uint32_t)(dev & 0xFF)) )


// minimal firmware version allowed for this library
#define FISHINO_FIRMWARE_VERSION_MIN VERSION(1,2,1)

#include <inttypes.h>

#include "IPAddress.h"

#include "Stream.h"

// SPI handshake I/O pin
#define ESP8266_HANDSHAKE 7

// SPI CS I/O pin
#define ESP8266_CS 10

typedef enum
{
	NULL_MODE,
	STATION_MODE,
	SOFTAP_MODE,
	STATIONAP_MODE

} WIFI_MODE;

typedef enum
{
    AUTH_OPEN           = 0,
    AUTH_WEP,
    AUTH_WPA_PSK,
    AUTH_WPA2_PSK,
    AUTH_WPA_WPA2_PSK,
    AUTH_MAX
    
} AUTH_MODE;

typedef enum
{
    STATION_IDLE = 0,
    STATION_CONNECTING,
    STATION_WRONG_PASSWORD,
    STATION_NO_AP_FOUND,
    STATION_CONNECT_FAIL,
    STATION_GOT_IP,
    STATION_TIMEOUT,
    STATION_BAD_SSID,

} JOIN_STATUS;

struct APINFO
{
	uint8_t authmode;
	char *ssid;
	int8_t rssi;
	uint8_t bssid[6];
	uint8_t channel;
};

typedef enum _SOCK_STATUS
{
	SOCK_INVALID,
	SOCK_DISCONNECTED,
	SOCK_CONNECTED,

} SOCK_STATUS;

typedef enum _PHY_MODE
{
	PHY_MODE_11B = 1,
	PHY_MODE_11G = 2,
	PHY_MODE_11N = 3
} PHY_MODE;

class FishinoClass
{
	private:
		
		// firmware version
		uint32_t _fwVersion;
		
		// SSID of joined AP, if any
		char *_ssid;

		// join status of last join command
		uint8_t _joinStatus;
		
		// scanned access point info
		APINFO *_apInfo;
		uint8_t _numAp;
		
		// frees APINFO array
		void freeApInfo(void);
		
		// find an APINFO given SSID
		APINFO const *findApInfo(const char *ssid);
		
		// low-level i/o routines
		void spiSelect(bool s);
		
		// wait for busy handshake, or 1 msecond
		// which comes first
		void spiWaitBusy(void);

		// check if data is available from spi
		bool spiDataAvailable(void);
		
		// wait for spi data available, timeout in milliseconds
		bool spiWaitDataAvailable(uint16_t tOut);
		
		// send/receive from 1 to 64 bytes
		// beware, ESP MUST know data length in advance
		bool spiSend(uint8_t const *b, uint8_t size);
		bool spiReceive(uint8_t *b, uint8_t size);

		// drops data from server (like spiReceive() but don't save it
		bool spiDrop(uint8_t size);
		
		// send stream terminator
		bool spiSendTerminator(void);
		
		// check stream terminator
		bool spiCheckTerminator(void);
		
		// send/receive following data size
		bool spiSendSize(uint16_t w);
		bool spiReceiveSize(uint16_t &w);
		
		// send/receive data bytes
		bool spiSendCommand(uint8_t cmd);
		bool spiSend8(uint8_t b);
		bool spiReceive8(uint8_t &b);
		bool spiSend16(uint16_t w);
		bool spiReceive16(uint16_t &w);
		bool spiSend32(uint32_t dw);
		bool spiReceive32(uint32_t &dw);
		
		// send/receive a string
		// received string is malloc-ed, MUST be freed with free()
		bool spiSendString(const char *str);
		bool spiReceiveString(char *&str);
		
		// send data by pointer and length
		bool spiSendData(uint8_t const *buf, uint16_t bufSize);
		
		// receive data, buffer allocated, must be freed by free()
		bool spiReceiveDataAlloc(uint8_t *&buf, uint16_t &bufSize);
		
		// receive data, buffer with enough space MUST be given
		// bufSize MUST be initialized with buffer size
		// bufSize is replaced with actual data size on success
		// on error or buffer overflow, bufSize is set to 0
		bool spiReceiveDataFixed(uint8_t *buf, uint16_t &bufSize);

		// low level TCP stuffs -- used by client and server
		
		// connect to host -- return socket number, or 0xff if error
		uint8_t connect(const char *host, uint16_t port);
		
		// connect to a secure host (HTTPS - SSL)
		uint8_t secureConnect(const char *host, uint16_t port);
		
		// disconnect from host -- return true on success, false otherwise
		bool disconnect(uint8_t sock);
		
		// write to socket -- return true on success, false otherwise
		bool write(uint8_t sock, uint8_t const *buf, uint16_t bufLen);
		
		// read from socket -- return true on success, false otherwise
		// buffer MUST be big enough to contain reqBytes
		// on exit, bufLen contains number of bytes actually read
		bool read(uint8_t sock, uint16_t reqBytes, uint8_t *buf, uint16_t &bufLen);

		// return number of available bytes on socket
		uint16_t available(uint8_t sock);

		// peek a data byte from socket, if any
		// returns true on success, false otherwise
		bool peek(uint8_t sock, uint8_t &b);
		
		// flush socket data
		bool flush(uint8_t sock);
		
		// connection status
		uint8_t status(uint8_t sock);

		// start a server on given port
		// ONLY ONE SERVER AT A TIME
		bool startServer(uint32_t port, uint16_t timeout);
		
		// stops server
		bool stopServer(void);
		
		// poll server for sockets with data
		// returns a dynamic allocated buffer which MUST be freed by free
		bool pollServer(bool all, uint8_t *&buf, uint16_t &count);
		
	public:

		// constructor
		FishinoClass();
		
		// destructor
		~FishinoClass();
		
		// reset ESP and wait for it to be ready
		// return true if ok, false if not ready
		bool reset(void);
		
		// set serial interface parameters
		bool setSerialParams(uint32_t speed, uint8_t parity, uint8_t bits, uint8_t stopBits);

		// Get firmware version
		uint32_t firmwareVersion();
		char *firmwareVersionStr();
		
		// set operation mode (one of WIFI_MODE)
		bool setMode(uint8_t mode);
		
		// get operation mode
		uint8_t getMode(void);

		// Start Wifi connection for OPEN networks
		// param ssid: Pointer to the SSID string.
		// return one of JOIN_STATUS result
		uint8_t begin(const char* ssid);

		// Start Wifi connection with encryption.
		// return one of JOIN_STATUS result
		uint8_t begin(const char* ssid, const char *passphrase);

		// setup station AP parameters without joining
		bool setStaConfig(const char *ssid);
		bool setStaConfig(const char *ssid, const char *passphrase);
		
		// get current station AP parameters
		// WARNING : returns dynamic buffers wich MUST be freed
		bool getStaConfig(char *&ssid, char *&pass);
		
		// join AP setup by setStaConfig call (or previous AP set by begin)
		uint8_t joinAp(void);
		
		// quits AP
		bool quitAp(void);

		// Change Ip configuration settings disabling the dhcp client
		//	param local_ip: 	Static ip configuration
		bool config(IPAddress local_ip);

		// Change Ip configuration settings disabling the dhcp client
		// param local_ip:			Static ip configuration
		// param dns_server:		IP configuration for DNS server
		// DNS SERVER SETTING NOT SUPPORTED BY NOW -- DUMMY
		bool config(IPAddress local_ip, IPAddress dns_server);

		// Change Ip configuration settings disabling the dhcp client
		// param local_ip:			Static ip configuration
		// param dns_server:		IP configuration for DNS server 1
		// param gateway :			Static gateway configuration
		// DNS SERVER SETTING NOT SUPPORTED BY NOW -- DUMMY
		bool config(IPAddress local_ip, IPAddress dns_server, IPAddress gateway);

		// Change Ip configuration settings disabling the dhcp client
		// param local_ip:			Static ip configuration
		// param dns_server:		IP configuration for DNS server 1
		// param gateway:			Static gateway configuration
		// param subnet:			Static Subnet mask
		// DNS SERVER SETTING NOT SUPPORTED BY NOW -- DUMMY
		bool config(IPAddress local_ip, IPAddress dns_server, IPAddress gateway, IPAddress subnet);

		// Change DNS Ip configuration
		// param dns_server1:		ip configuration for DNS server 1
		// DNS SERVER SETTING NOT SUPPORTED BY NOW -- DUMMY
		bool setDNS(IPAddress dns_server1);

		// Change DNS Ip configuration
		// param dns_server1:		ip configuration for DNS server 1
		// param dns_server2:		ip configuration for DNS server 2
		// DNS SERVER SETTING NOT SUPPORTED BY NOW -- DUMMY
		bool setDNS(IPAddress dns_server1, IPAddress dns_server2);

		// Disconnect from the network
		// return: one value of wl_status_t enum
		bool disconnect(void);

		// Get the interface MAC address.
		// return: pointer to A STATIC uint8_t array with length WL_MAC_ADDR_LENGTH
		const uint8_t* macAddress(void);

		// Get the interface IP address.
		// return: Ip address value
		IPAddress localIP();

		// Get the interface subnet mask address.
		// return: subnet mask address value
		IPAddress subnetMask();

		// Get the gateway ip address.
		// return: gateway ip address value
		IPAddress gatewayIP();

		// Return the current SSID associated with the network
		// return: ssid string
		const char* SSID();

		// Return the current BSSID associated with the network.
		// It is the MAC address of the Access Point
		// return: pointer to uint8_t array with length WL_MAC_ADDR_LENGTH
		const uint8_t* BSSID();

		// Return the current RSSI /Received Signal Strength in dBm)
		// associated with the network
		// return: signed value
		int32_t RSSI();

		// Return the Encryption Type associated with the network
		// return: one value of wl_enc_type enum
		uint8_t	encryptionType();

		// Start scan WiFi networks available
		// return: Number of discovered networks
		uint8_t scanNetworks();

		// Return the SSID discovered during the network scan.
		// param networkItem: specify from which network item want to get the information
		// return: ssid string of the specified item on the networks scanned list
		const char* SSID(uint8_t networkItem);

		// Return the encryption type of the networks discovered during the scanNetworks
		// param networkItem: specify from which network item want to get the information
		// return: encryption type (enum wl_enc_type) of the specified item on the networks scanned list
		uint8_t encryptionType(uint8_t networkItem);

		// Return the RSSI of the networks discovered during the scanNetworks
		// param networkItem: specify from which network item want to get the information
		// return: signed value of RSSI of the specified item on the networks scanned list
		int32_t RSSI(uint8_t networkItem);

		// Return Connection status.
		// return: one of the value defined in wl_status_t
		uint8_t status();
		
		// "uniform" functions to set station and AP parameters
		// station ones replicates the above ones, jus to have meaningful names
		
		bool setStaIP(IPAddress ip);
		bool setStaMAC(uint8_t const *mac);
		bool setStaGateway(IPAddress gw);
		bool setStaNetMask(IPAddress nm);
		
		// station DHCP client
		bool staStartDHCP(void);
		bool staStopDHCP(void);
		bool getStaDHCPStatus(void);

		bool setApIP(IPAddress ip);
		bool setApMAC(uint8_t const *mac);
		bool setApGateway(IPAddress gw);
		bool setApNetMask(IPAddress nm);

		bool setApIPInfo(IPAddress ip, IPAddress gateway, IPAddress netmask);

		// Resolve the given hostname to an IP address.
		// param aHostname: Name to be resolved
		// param aResult: IPAddress structure to store the returned IP address
		// result: 1 if aIPAddrString was successfully converted to an IP address,
		// else error code
		bool hostByName(const char* aHostname, IPAddress& aResult);

		// softAp DHCP
		bool softApStartDHCPServer(void);
		bool softApStartDHCPServer(IPAddress startIP, IPAddress endIP);
		bool softApStopDHCPServer(void);
		bool getSoftApDHCPServerStatus(void);

		// softAp configuration
		
		// get ap SSID - return a dynamic buffer that MUST be freed by free()
		// return NULL on error
		char *softApGetSSID(void);
		
		// get ap PASSWORD - return a dynamic buffer that MUST be freed by free()
		// return NULL on error
		char *softApGetPassword(void);
		
		// return ap channel
		// return 0 on error
		uint8_t softApGetChannel(void);
		
		// return ap hidden state
		bool softApGetHidden(void);
		
		// get all softAp config data
		// warning - all returned strings MUST be freed by free()
		bool softApGetConfig(char *&SSID, char *&pass, uint8_t &channel, bool &hidden);
		
		// set softAp parameters
		bool softApConfig(const char *SSID, const char *pass, uint8_t channel, bool hidden = false);

		friend class FishinoClient;
		friend class FishinoSecureClient;
		friend class FishinoServer;

		// extra I/O pins on ESP module

		// pinMode for ESP I/O
		bool pinMode(uint8_t pin, uint8_t mode);

		// digital I/O for ESP I/O
		uint8_t digitalRead(uint8_t pin);
		bool digitalWrite(uint8_t pin, uint8_t val);
		
		// analog read
		uint16_t analogRead(void);
		
		// physical mode handling
		uint8_t getPhyMode(void);
		bool setPhyMode(uint8_t mode);
		
};

extern FishinoClass Fishino;

#include "FishinoClient.h"
#include "FishinoServer.h"

#endif
